@@ -24,9 +24,11 @@ use crate::error::CoreError::ReadFileSliceError;
 use crate::expr::filter::{Filter, SchemableFilter};
 use crate::file_group::file_slice::FileSlice;
 use crate::file_group::log_file::scanner::LogFileScanner;
+use crate::file_group::record_batches::RecordBatches;
 use crate::merge::record_merger::RecordMerger;
 use crate::metadata::meta_field::MetaField;
 use crate::storage::Storage;
+use crate::table::builder::OptionResolver;
 use crate::timeline::selector::InstantRange;
 use crate::Result;
 use arrow::compute::and;
@@ -44,16 +46,20 @@ pub struct FileGroupReader {
 }
 
 impl FileGroupReader {
-    pub(crate) fn new_with_configs_and_options<I, K, V>(
+    /// Creates a new reader with the given Hudi configurations and overwriting options.
+    ///
+    /// # Notes
+    /// This API does **not** use [`OptionResolver`] that loads table properties from storage to resolve options.
+    pub(crate) fn new_with_configs_and_overwriting_options<I, K, V>(
         hudi_configs: Arc<HudiConfigs>,
-        options: I,
+        overwriting_options: I,
     ) -> Result<Self>
     where
         I: IntoIterator<Item = (K, V)>,
         K: AsRef<str>,
         V: Into<String>,
     {
-        let (hudi_opts, others) = split_hudi_options_from_others(options);
+        let (hudi_opts, others) = split_hudi_options_from_others(overwriting_options);
 
         let mut final_opts = hudi_configs.as_options();
         final_opts.extend(hudi_opts);
@@ -71,18 +77,30 @@ impl FileGroupReader {
     /// # Arguments
     ///     * `base_uri` - The base URI of the file group's residing table.
     ///     * `options` - Additional options for the reader.
+    ///
+    /// # Notes
+    /// This API uses [`OptionResolver`] that loads table properties from storage to resolve options.
     pub fn new_with_options<I, K, V>(base_uri: &str, options: I) -> Result<Self>
     where
         I: IntoIterator<Item = (K, V)>,
         K: AsRef<str>,
         V: Into<String>,
     {
-        let hudi_configs = Arc::new(HudiConfigs::new([(
-            HudiTableConfig::BasePath.as_ref().to_string(),
-            base_uri.to_string(),
-        )]));
+        tokio::runtime::Builder::new_current_thread()
+            .enable_all()
+            .build()?
+            .block_on(async {
+                let mut resolver = OptionResolver::new_with_options(base_uri, options);
+                resolver.resolve_options().await?;
+                let hudi_configs = Arc::new(HudiConfigs::new(resolver.hudi_options));
+                let storage =
+                    Storage::new(Arc::new(resolver.storage_options), hudi_configs.clone())?;
 
-        Self::new_with_configs_and_options(hudi_configs, options)
+                Ok(Self {
+                    hudi_configs,
+                    storage,
+                })
+            })
     }
 
     fn create_filtering_mask_for_base_file_records(
@@ -168,6 +186,17 @@ impl FileGroupReader {
         }
     }
 
+    /// Same as [FileGroupReader::read_file_slice_by_base_file_path], but blocking.
+    pub fn read_file_slice_by_base_file_path_blocking(
+        &self,
+        relative_path: &str,
+    ) -> Result<RecordBatch> {
+        tokio::runtime::Builder::new_current_thread()
+            .enable_all()
+            .build()?
+            .block_on(self.read_file_slice_by_base_file_path(relative_path))
+    }
+
     fn create_instant_range_for_log_file_scan(&self) -> InstantRange {
         let timezone = self
             .hudi_configs
@@ -201,30 +230,39 @@ impl FileGroupReader {
         if base_file_only {
             self.read_file_slice_by_base_file_path(&relative_path).await
         } else {
-            let base_record_batch = self
-                .read_file_slice_by_base_file_path(&relative_path)
-                .await?;
-            let schema = base_record_batch.schema();
-            let mut all_record_batches = vec![base_record_batch];
-
             let log_file_paths = file_slice
                 .log_files
                 .iter()
                 .map(|log_file| file_slice.log_file_relative_path(log_file))
                 .collect::<Result<Vec<String>>>()?;
             let instant_range = self.create_instant_range_for_log_file_scan();
-            let log_record_batches =
-                LogFileScanner::new(self.hudi_configs.clone(), self.storage.clone())
-                    .scan(log_file_paths, &instant_range)
-                    .await?;
-            for log_record_batch in log_record_batches {
-                all_record_batches.extend(log_record_batch);
-            }
+            let log_batches = LogFileScanner::new(self.hudi_configs.clone(), self.storage.clone())
+                .scan(log_file_paths, &instant_range)
+                .await?;
 
-            let merger = RecordMerger::new(self.hudi_configs.clone());
-            merger.merge_record_batches(&schema, &all_record_batches)
+            let base_batch = self
+                .read_file_slice_by_base_file_path(&relative_path)
+                .await?;
+            let schema = base_batch.schema();
+            let num_data_batches = log_batches.num_data_batches() + 1;
+            let num_delete_batches = log_batches.num_delete_batches();
+            let mut all_batches =
+                RecordBatches::new_with_capacity(num_data_batches, num_delete_batches);
+            all_batches.push_data_batch(base_batch);
+            all_batches.extend(log_batches);
+
+            let merger = RecordMerger::new(schema.clone(), self.hudi_configs.clone());
+            merger.merge_record_batches(all_batches)
         }
     }
+
+    /// Same as [FileGroupReader::read_file_slice], but blocking.
+    pub fn read_file_slice_blocking(&self, file_slice: &FileSlice) -> Result<RecordBatch> {
+        tokio::runtime::Builder::new_current_thread()
+            .enable_all()
+            .build()?
+            .block_on(self.read_file_slice(file_slice))
+    }
 }
 
 #[cfg(test)]
@@ -232,32 +270,79 @@ mod tests {
     use super::*;
     use crate::config::util::empty_options;
     use crate::error::CoreError;
+    use crate::Result;
     use arrow::array::{ArrayRef, Int64Array, StringArray};
     use arrow::record_batch::RecordBatch;
     use arrow_schema::{DataType, Field, Schema};
+    use std::fs::canonicalize;
+    use std::path::PathBuf;
     use std::sync::Arc;
+    use url::Url;
+
+    fn get_non_existent_base_uri() -> String {
+        "file:///non-existent-path/table".to_string()
+    }
+
+    fn get_base_uri_with_valid_props() -> String {
+        let url = Url::from_file_path(
+            canonicalize(
+                PathBuf::from("tests")
+                    .join("data")
+                    .join("table_props_valid"),
+            )
+            .unwrap(),
+        )
+        .unwrap();
+        url.as_ref().to_string()
+    }
+
+    fn get_base_uri_with_valid_props_minimum() -> String {
+        let url = Url::from_file_path(
+            canonicalize(
+                PathBuf::from("tests")
+                    .join("data")
+                    .join("table_props_valid_minimum"),
+            )
+            .unwrap(),
+        )
+        .unwrap();
+        url.as_ref().to_string()
+    }
+
+    fn get_base_uri_with_invalid_props() -> String {
+        let url = Url::from_file_path(
+            canonicalize(
+                PathBuf::from("tests")
+                    .join("data")
+                    .join("table_props_invalid"),
+            )
+            .unwrap(),
+        )
+        .unwrap();
+        url.as_ref().to_string()
+    }
 
     #[test]
-    fn test_new_with_options() -> Result<()> {
+    fn test_new_with_options() {
         let options = vec![("key1", "value1"), ("key2", "value2")];
-        let reader = FileGroupReader::new_with_options("/tmp/hudi_data", options)?;
+        let base_uri = get_base_uri_with_valid_props();
+        let reader = FileGroupReader::new_with_options(&base_uri, options).unwrap();
         assert!(!reader.storage.options.is_empty());
         assert!(reader
             .storage
             .hudi_configs
             .contains(HudiTableConfig::BasePath));
-        Ok(())
     }
 
-    #[tokio::test]
-    async fn test_read_file_slice_returns_error() {
-        let reader =
-            FileGroupReader::new_with_options("file:///non-existent-path/table", empty_options())
-                .unwrap();
-        let result = reader
-            .read_file_slice_by_base_file_path("non_existent_file")
-            .await;
-        assert!(matches!(result.unwrap_err(), ReadFileSliceError(_)));
+    #[test]
+    fn test_new_with_options_invalid_base_uri_or_invalid_props() {
+        let base_uri = get_non_existent_base_uri();
+        let result = FileGroupReader::new_with_options(&base_uri, empty_options());
+        assert!(result.is_err());
+
+        let base_uri = get_base_uri_with_invalid_props();
+        let result = FileGroupReader::new_with_options(&base_uri, empty_options());
+        assert!(result.is_err())
     }
 
     fn create_test_record_batch() -> Result<RecordBatch> {
@@ -279,11 +364,12 @@ mod tests {
 
     #[test]
     fn test_create_filtering_mask_for_base_file_records() -> Result<()> {
-        let base_uri = "file:///non-existent-path/table";
+        let base_uri = get_base_uri_with_valid_props_minimum();
         let records = create_test_record_batch()?;
-        // Test case 1: No meta fields populated
+
+        // Test case 1: Disable populating the meta fields
         let reader = FileGroupReader::new_with_options(
-            base_uri,
+            &base_uri,
             [
                 (HudiTableConfig::PopulatesMetaFields.as_ref(), "false"),
                 (HudiReadConfig::FileGroupStartTimestamp.as_ref(), "2"),
@@ -293,13 +379,13 @@ mod tests {
         assert_eq!(mask, None, "Commit time filtering should not be needed");
 
         // Test case 2: No commit time filtering options
-        let reader = FileGroupReader::new_with_options(base_uri, empty_options())?;
+        let reader = FileGroupReader::new_with_options(&base_uri, empty_options())?;
         let mask = reader.create_filtering_mask_for_base_file_records(&records)?;
         assert_eq!(mask, None);
 
         // Test case 3: Filtering commit time > '2'
         let reader = FileGroupReader::new_with_options(
-            base_uri,
+            &base_uri,
             [(HudiReadConfig::FileGroupStartTimestamp, "2")],
         )?;
         let mask = reader.create_filtering_mask_for_base_file_records(&records)?;
@@ -311,7 +397,7 @@ mod tests {
 
         // Test case 4: Filtering commit time <= '4'
         let reader = FileGroupReader::new_with_options(
-            base_uri,
+            &base_uri,
             [(HudiReadConfig::FileGroupEndTimestamp, "4")],
         )?;
         let mask = reader.create_filtering_mask_for_base_file_records(&records)?;
@@ -319,7 +405,7 @@ mod tests {
 
         // Test case 5: Filtering commit time > '2' and <= '4'
         let reader = FileGroupReader::new_with_options(
-            base_uri,
+            &base_uri,
             [
                 (HudiReadConfig::FileGroupStartTimestamp, "2"),
                 (HudiReadConfig::FileGroupEndTimestamp, "4"),
