@@ -43,12 +43,15 @@
 //! }
 //! ```
 
+mod avro_to_arrow;
 pub mod config;
 pub mod error;
 pub mod expr;
 pub mod file_group;
 pub mod merge;
 pub mod metadata;
+mod record;
+pub mod schema;
 pub mod storage;
 pub mod table;
 pub mod timeline;
