@@ -20,34 +20,33 @@
 use paste::paste;
 use std::collections::HashMap;
 use std::env;
-use std::hash::Hash;
 use std::path::PathBuf;
 use std::sync::Arc;
-use strum::IntoEnumIterator;
 
-use crate::config::internal::HudiInternalConfig::SkipConfigValidation;
-use crate::config::read::HudiReadConfig;
 use crate::config::table::HudiTableConfig;
-use crate::config::table::HudiTableConfig::{
-    DropsPartitionFields, TableVersion, TimelineLayoutVersion,
-};
 use crate::config::util::{parse_data_for_options, split_hudi_options_from_others};
 use crate::config::{HudiConfigs, HUDI_CONF_DIR};
-use crate::error::CoreError;
-use crate::merge::record_merger::RecordMerger;
 use crate::storage::Storage;
 use crate::table::fs_view::FileSystemView;
+use crate::table::validation::validate_configs;
 use crate::table::Table;
 use crate::timeline::Timeline;
+use crate::util::collection::extend_if_absent;
 use crate::Result;
 
 /// Builder for creating a [Table] instance.
 #[derive(Debug, Clone)]
 pub struct TableBuilder {
-    base_uri: String,
-    hudi_options: HashMap<String, String>,
-    storage_options: HashMap<String, String>,
-    options: HashMap<String, String>,
+    option_resolver: OptionResolver,
+}
+
+/// Resolver for options including Hudi options, storage options, and generic options.
+#[derive(Debug, Clone)]
+pub struct OptionResolver {
+    pub base_uri: String,
+    pub hudi_options: HashMap<String, String>,
+    pub storage_options: HashMap<String, String>,
+    pub options: HashMap<String, String>,
 }
 
 macro_rules! impl_with_options {
@@ -62,7 +61,8 @@ macro_rules! impl_with_options {
                         K: AsRef<str>,
                         V: Into<String>,
                     {
-                        self.$field.insert(k.as_ref().to_string(), v.into());
+                        let option_resolver = &mut self.option_resolver;
+                        option_resolver.$field.insert(k.as_ref().to_string(), v.into());
                         self
                     }
 
@@ -74,7 +74,8 @@ macro_rules! impl_with_options {
                         K: AsRef<str>,
                         V: Into<String>,
                     {
-                        self.$field.extend(options.into_iter().map(|(k, v)| (k.as_ref().to_string(), v.into())));
+                        let option_resolver = &mut self.option_resolver;
+                        option_resolver.$field.extend(options.into_iter().map(|(k, v)| (k.as_ref().to_string(), v.into())));
                         self
                     }
                 }
@@ -96,23 +97,18 @@ impl_with_options!(
 impl TableBuilder {
     /// Create Hudi table builder from base table uri
     pub fn from_base_uri(base_uri: &str) -> Self {
-        TableBuilder {
-            base_uri: base_uri.to_string(),
-            storage_options: HashMap::new(),
-            hudi_options: HashMap::new(),
-            options: HashMap::new(),
-        }
+        let option_resolver = OptionResolver::new(base_uri);
+        TableBuilder { option_resolver }
     }
 
     pub async fn build(&mut self) -> Result<Table> {
-        self.resolve_options().await?;
+        let option_resolver = &mut self.option_resolver;
 
-        let hudi_configs = HudiConfigs::new(self.hudi_options.iter());
+        option_resolver.resolve_options().await?;
 
-        Self::validate_configs(&hudi_configs).expect("Hudi configs are not valid.");
+        let hudi_configs = Arc::from(HudiConfigs::new(option_resolver.hudi_options.iter()));
 
-        let hudi_configs = Arc::from(hudi_configs);
-        let storage_options = Arc::from(self.storage_options.clone());
+        let storage_options = Arc::from(self.option_resolver.storage_options.clone());
 
         let timeline =
             Timeline::new_from_storage(hudi_configs.clone(), storage_options.clone()).await?;
@@ -127,6 +123,37 @@ impl TableBuilder {
             file_system_view,
         })
     }
+}
+
+impl OptionResolver {
+    /// Create a new [OptionResolver] with the given base URI.
+    pub fn new(base_uri: &str) -> Self {
+        Self {
+            base_uri: base_uri.to_string(),
+            hudi_options: HashMap::new(),
+            storage_options: HashMap::new(),
+            options: HashMap::new(),
+        }
+    }
+
+    /// Create a new [OptionResolver] with the given base URI and options.
+    pub fn new_with_options<I, K, V>(base_uri: &str, options: I) -> Self
+    where
+        I: IntoIterator<Item = (K, V)>,
+        K: AsRef<str>,
+        V: Into<String>,
+    {
+        let options = options
+            .into_iter()
+            .map(|(k, v)| (k.as_ref().to_string(), v.into()))
+            .collect();
+        Self {
+            base_uri: base_uri.to_string(),
+            hudi_options: HashMap::new(),
+            storage_options: HashMap::new(),
+            options,
+        }
+    }
 
     /// Resolve all options by combining the ones from hoodie.properties, user-provided options,
     /// env vars, and global Hudi configs. The precedence order is as follows:
@@ -138,16 +165,20 @@ impl TableBuilder {
     /// 5. Global Hudi configs
     ///
     /// [note] Error may occur when 1 and 2 have conflicts.
-    async fn resolve_options(&mut self) -> Result<()> {
+    pub async fn resolve_options(&mut self) -> Result<()> {
         self.resolve_user_provided_options();
 
-        // if any user-provided options are intended for cloud storage and in uppercase,
+        // If any user-provided options are intended for cloud storage and in uppercase,
         // convert them to lowercase. This is to allow `object_store` to pick them up.
         self.resolve_cloud_env_vars();
 
         // At this point, we have resolved the storage options needed for accessing the storage layer.
         // We can now resolve the hudi options
-        self.resolve_hudi_options().await
+        self.resolve_hudi_options().await?;
+
+        // Validate the resolved Hudi options
+        let hudi_configs = HudiConfigs::new(self.hudi_options.iter());
+        validate_configs(&hudi_configs)
     }
 
     fn resolve_user_provided_options(&mut self) {
@@ -162,8 +193,8 @@ impl TableBuilder {
 
         // Combine generic options (lower precedence) with explicit options.
         // Note that we treat all non-Hudi options as storage options
-        Self::extend_if_absent(&mut self.hudi_options, &generic_hudi_opts);
-        Self::extend_if_absent(&mut self.storage_options, &generic_other_opts)
+        extend_if_absent(&mut self.hudi_options, &generic_hudi_opts);
+        extend_if_absent(&mut self.storage_options, &generic_other_opts)
     }
 
     fn resolve_cloud_env_vars(&mut self) {
@@ -202,8 +233,9 @@ impl TableBuilder {
         let bytes = storage.get_file_data(".hoodie/hoodie.properties").await?;
         let table_properties = parse_data_for_options(&bytes, "=")?;
 
-        // We currently treat all table properties as the highest precedence, which is valid for most cases.
-        // TODO: handle the case where the same key is present in both table properties and options
+        // Table properties on storage (hoodie.properties) should have the highest precedence,
+        // except for writer-changeable properties like enabling metadata table/indexes.
+        // TODO: return err when user-provided options conflict with table properties
         for (k, v) in table_properties {
             options.insert(k.to_string(), v.to_string());
         }
@@ -235,62 +267,6 @@ impl TableBuilder {
 
         Ok(())
     }
-
-    fn validate_configs(hudi_configs: &HudiConfigs) -> Result<()> {
-        if hudi_configs
-            .get_or_default(SkipConfigValidation)
-            .to::<bool>()
-        {
-            return Ok(());
-        }
-
-        for conf in HudiTableConfig::iter() {
-            hudi_configs.validate(conf)?
-        }
-
-        for conf in HudiReadConfig::iter() {
-            hudi_configs.validate(conf)?
-        }
-
-        // additional validation
-        let table_version = hudi_configs.get(TableVersion)?.to::<isize>();
-        if !(5..=6).contains(&table_version) {
-            return Err(CoreError::Unsupported(
-                "Only support table version 5 and 6.".to_string(),
-            ));
-        }
-
-        let timeline_layout_version = hudi_configs.get(TimelineLayoutVersion)?.to::<isize>();
-        if timeline_layout_version != 1 {
-            return Err(CoreError::Unsupported(
-                "Only support timeline layout version 1.".to_string(),
-            ));
-        }
-
-        let drops_partition_cols = hudi_configs
-            .get_or_default(DropsPartitionFields)
-            .to::<bool>();
-        if drops_partition_cols {
-            return Err(CoreError::Unsupported(format!(
-                "Only support when `{}` is disabled",
-                DropsPartitionFields.as_ref()
-            )));
-        }
-
-        RecordMerger::validate_configs(hudi_configs)?;
-
-        Ok(())
-    }
-
-    fn extend_if_absent<K, V>(target: &mut HashMap<K, V>, source: &HashMap<K, V>)
-    where
-        K: Eq + Hash + Clone,
-        V: Clone,
-    {
-        for (key, value) in source {
-            target.entry(key.clone()).or_insert_with(|| value.clone());
-        }
-    }
 }
 
 #[cfg(test)]
@@ -298,19 +274,15 @@ mod tests {
     use super::*;
 
     fn create_table_builder() -> TableBuilder {
-        TableBuilder {
-            base_uri: "test_uri".to_string(),
-            hudi_options: HashMap::new(),
-            storage_options: HashMap::new(),
-            options: HashMap::new(),
-        }
+        let option_resolver = OptionResolver::new("test_uri");
+        TableBuilder { option_resolver }
     }
 
     #[test]
     fn test_with_hudi_option() {
         let builder = create_table_builder();
 
-        let updated = builder.with_hudi_option("key", "value");
+        let updated = builder.with_hudi_option("key", "value").option_resolver;
         assert_eq!(updated.hudi_options["key"], "value")
     }
 
@@ -319,7 +291,7 @@ mod tests {
         let builder = create_table_builder();
 
         let options = [("key1", "value1"), ("key2", "value2")];
-        let updated = builder.with_hudi_options(options);
+        let updated = builder.with_hudi_options(options).option_resolver;
         assert_eq!(updated.hudi_options["key1"], "value1");
         assert_eq!(updated.hudi_options["key2"], "value2")
     }
@@ -328,7 +300,7 @@ mod tests {
     fn test_with_storage_option() {
         let builder = create_table_builder();
 
-        let updated = builder.with_storage_option("key", "value");
+        let updated = builder.with_storage_option("key", "value").option_resolver;
         assert_eq!(updated.storage_options["key"], "value")
     }
 
@@ -337,7 +309,7 @@ mod tests {
         let builder = create_table_builder();
 
         let options = [("key1", "value1"), ("key2", "value2")];
-        let updated = builder.with_storage_options(options);
+        let updated = builder.with_storage_options(options).option_resolver;
         assert_eq!(updated.storage_options["key1"], "value1");
         assert_eq!(updated.storage_options["key2"], "value2");
     }
@@ -346,7 +318,7 @@ mod tests {
     fn test_with_option() {
         let builder = create_table_builder();
 
-        let updated = builder.with_option("key", "value");
+        let updated = builder.with_option("key", "value").option_resolver;
         assert_eq!(updated.options["key"], "value")
     }
 
@@ -355,7 +327,7 @@ mod tests {
         let builder = create_table_builder();
 
         let options = [("key1", "value1"), ("key2", "value2")];
-        let updated = builder.with_options(options);
+        let updated = builder.with_options(options).option_resolver;
         assert_eq!(updated.options["key1"], "value1");
         assert_eq!(updated.options["key2"], "value2")
     }
@@ -381,17 +353,18 @@ mod tests {
                 ("hoodie.option2", "value2-1"),
             ]);
 
-        builder.resolve_user_provided_options();
+        let resolver = &mut builder.option_resolver;
+        resolver.resolve_user_provided_options();
 
-        assert_eq!(builder.hudi_options.len(), 4);
-        assert_eq!(builder.hudi_options["hoodie.base.path"], "/tmp/hudi_data");
-        assert_eq!(builder.hudi_options["hoodie.option1"], "value1-2");
-        assert_eq!(builder.hudi_options["hoodie.option2"], "value2-1");
-        assert_eq!(builder.hudi_options["hoodie.option3"], "value3");
-        assert_eq!(builder.storage_options.len(), 2);
-        assert_eq!(builder.storage_options["AWS_REGION"], "us-east-1");
+        assert_eq!(resolver.hudi_options.len(), 4);
+        assert_eq!(resolver.hudi_options["hoodie.base.path"], "/tmp/hudi_data");
+        assert_eq!(resolver.hudi_options["hoodie.option1"], "value1-2");
+        assert_eq!(resolver.hudi_options["hoodie.option2"], "value2-1");
+        assert_eq!(resolver.hudi_options["hoodie.option3"], "value3");
+        assert_eq!(resolver.storage_options.len(), 2);
+        assert_eq!(resolver.storage_options["AWS_REGION"], "us-east-1");
         assert_eq!(
-            builder.storage_options["AWS_ENDPOINT"],
+            resolver.storage_options["AWS_ENDPOINT"],
             "s3.us-east-1.amazonaws.com"
         );
     }
