@@ -30,15 +30,16 @@ use async_trait::async_trait;
 use datafusion::catalog::{Session, TableProviderFactory};
 use datafusion::datasource::listing::PartitionedFile;
 use datafusion::datasource::object_store::ObjectStoreUrl;
-use datafusion::datasource::physical_plan::parquet::ParquetExecBuilder;
+use datafusion::datasource::physical_plan::parquet::source::ParquetSource;
 use datafusion::datasource::physical_plan::FileScanConfig;
+use datafusion::datasource::source::DataSourceExec;
 use datafusion::datasource::TableProvider;
+use datafusion::error::Result;
 use datafusion::logical_expr::Operator;
 use datafusion::physical_plan::ExecutionPlan;
 use datafusion_common::config::TableParquetOptions;
 use datafusion_common::DFSchema;
 use datafusion_common::DataFusionError::Execution;
-use datafusion_common::Result;
 use datafusion_expr::{CreateExternalTable, Expr, TableProviderFilterPushDown, TableType};
 use datafusion_physical_expr::create_physical_expr;
 
@@ -47,7 +48,6 @@ use hudi_core::config::read::HudiReadConfig::InputPartitions;
 use hudi_core::config::util::empty_options;
 use hudi_core::storage::util::{get_scheme_authority, join_url_segments};
 use hudi_core::table::Table as HudiTable;
-use hudi_core::util::StrTupleRef;
 
 /// Create a `HudiDataSource`.
 /// Used for Datafusion to query Hudi tables
@@ -179,7 +179,7 @@ impl TableProvider for HudiDataSource {
         let pushdown_filters = exprs_to_filters(filters);
         let file_slices = self
             .table
-            .get_file_slices_splits(self.get_input_partitions(), &pushdown_filters.as_strs())
+            .get_file_slices_splits(self.get_input_partitions(), pushdown_filters)
             .await
             .map_err(|e| Execution(format!("Failed to get file slices from Hudi table: {}", e)))?;
         let base_url = self.table.base_url();
@@ -203,26 +203,29 @@ impl TableProvider for HudiDataSource {
 
         let base_url = self.table.base_url();
         let url = ObjectStoreUrl::parse(get_scheme_authority(&base_url))?;
-        let fsc = FileScanConfig::new(url, self.schema())
-            .with_file_groups(parquet_file_groups)
-            .with_projection(projection.cloned())
-            .with_limit(limit);
 
         let parquet_opts = TableParquetOptions {
             global: state.config_options().execution.parquet.clone(),
             column_specific_options: Default::default(),
             key_value_metadata: Default::default(),
         };
-        let mut exec_builder = ParquetExecBuilder::new_with_options(fsc, parquet_opts);
-
+        let table_schema = self.schema();
+        let mut parquet_source = ParquetSource::new(parquet_opts);
         let filter = filters.iter().cloned().reduce(|acc, new| acc.and(new));
         if let Some(expr) = filter {
-            let df_schema = DFSchema::try_from(self.schema())?;
+            let df_schema = DFSchema::try_from(table_schema.clone())?;
             let predicate = create_physical_expr(&expr, &df_schema, state.execution_props())?;
-            exec_builder = exec_builder.with_predicate(predicate)
+            parquet_source = parquet_source.with_predicate(table_schema.clone(), predicate)
         }
 
-        Ok(exec_builder.build_arc())
+        let fsc = Arc::new(
+            FileScanConfig::new(url, table_schema, Arc::new(parquet_source))
+                .with_file_groups(parquet_file_groups)
+                .with_projection(projection.cloned())
+                .with_limit(limit),
+        );
+
+        Ok(Arc::new(DataSourceExec::new(fsc)))
     }
 
     fn supports_filters_pushdown(
@@ -338,6 +341,8 @@ mod tests {
 
     use datafusion::logical_expr::BinaryExpr;
     use hudi_core::config::read::HudiReadConfig::InputPartitions;
+    use hudi_core::metadata::meta_field::MetaField;
+    use hudi_test::assert_arrow_field_names_eq;
     use hudi_test::SampleTable::{
         V6ComplexkeygenHivestyle, V6Empty, V6Nonpartitioned, V6SimplekeygenHivestyleNoMetafields,
         V6SimplekeygenNonhivestyle, V6SimplekeygenNonhivestyleOverwritetable,
@@ -359,13 +364,16 @@ mod tests {
     }
 
     #[tokio::test]
-    async fn test_get_empty_schema_from_empty_table() {
+    async fn test_get_create_schema_from_empty_table() {
         let table_provider =
             HudiDataSource::new_with_options(V6Empty.path_to_cow().as_str(), empty_options())
                 .await
                 .unwrap();
         let schema = table_provider.schema();
-        assert!(schema.fields().is_empty());
+        assert_arrow_field_names_eq!(
+            schema,
+            [MetaField::field_names(), vec!["id", "name", "isActive"]].concat()
+        );
     }
 
     async fn register_test_table_with_session<I, K, V>(
@@ -463,7 +471,7 @@ mod tests {
             table_name
         )));
         assert!(plan_lines[4].starts_with(
-            "FilterExec: CAST(id@0 AS Int64) % 2 = 0 AND get_field(structField@3, field2) > 30"
+            "FilterExec: CAST(id@0 AS Int64) % 2 = 0 AND name@1 != Alice AND get_field(structField@3, field2) > 30"
         ));
         assert!(plan_lines[5].contains(&format!("input_partitions={}", planned_input_partitioned)));
     }
@@ -482,7 +490,7 @@ mod tests {
     }
 
     #[tokio::test]
-    async fn test_datafusion_read_hudi_table() {
+    async fn test_datafusion_read_hudi_table_with_partition_filter_pushdown() {
         for (test_table, use_sql, planned_input_partitions) in &[
             (V6ComplexkeygenHivestyle, true, 2),
             (V6Nonpartitioned, true, 1),
@@ -499,7 +507,7 @@ mod tests {
             let sql = format!(
                 r#"
             SELECT id, name, isActive, structField.field2
-            FROM {} WHERE id % 2 = 0
+            FROM {} WHERE id % 2 = 0 AND name != 'Alice'
             AND structField.field2 > 30 ORDER BY name LIMIT 10"#,
                 test_table.as_ref()
             );
@@ -523,7 +531,7 @@ mod tests {
     }
 
     #[tokio::test]
-    async fn test_datafusion_read_hudi_table_with_replacecommits() {
+    async fn test_datafusion_read_hudi_table_with_replacecommits_with_partition_filter_pushdown() {
         for (test_table, use_sql, planned_input_partitions) in
             &[(V6SimplekeygenNonhivestyleOverwritetable, true, 1)]
         {
@@ -536,7 +544,7 @@ mod tests {
             let sql = format!(
                 r#"
             SELECT id, name, isActive, structField.field2
-            FROM {} WHERE id % 2 = 0
+            FROM {} WHERE id % 2 = 0 AND name != 'Alice'
             AND structField.field2 > 30 ORDER BY name LIMIT 10"#,
                 test_table.as_ref()
             );
