@@ -16,8 +16,8 @@
 #  under the License.
 
 import os
+import pathlib
 import zipfile
-from pathlib import Path
 
 import pytest
 
@@ -34,7 +34,7 @@ def _extract_testing_table(zip_file_path, target_path) -> str:
     ]
 )
 def get_sample_table(request, tmp_path) -> str:
-    fixture_path = "tests/table"
+    fixture_path = pathlib.Path(__file__).parent.joinpath("table")
     table_name = request.param
-    zip_file_path = Path(fixture_path).joinpath(f"{table_name}.zip")
+    zip_file_path = pathlib.Path(fixture_path).joinpath(f"{table_name}.zip")
     return _extract_testing_table(zip_file_path, tmp_path)
